/*
 *  Spot and Texture OpenGL Demonstrator
 *  Moving parallel or spot light with textures and ambient light control
 *
 *  Created by brisson on 2012/05/02.
 *  Updated by brisson on 2013/05/04.
 *  Copyright (c) 2012-2013 J M. Brisson Lopes & Instituto Superior Tecnico. All rights reserved.
 *
 */

#include "SpotAndTexture.h"

#if defined(__APPLE__) || defined(MACOSX)
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif

#include <cstdlib>
#include <string>
#include <sstream>
#include <iostream>
#include <algorithm>
#include "BMPReader.h"

using std::string;
using std::ostringstream;
using std::ios;
using std::min;
using std::max;

// ObjectManager ------------------------------------------------------

ObjectManager::ObjectManager()
{
    
    DrawSphere = false;
    Sphere = glGenLists( 1);
    float sphereDiff[] = { .78, .57, .11, 1.};
    float sphereSpec[] = { .99, .94, .81, 1.};
    glNewList( Sphere, GL_COMPILE);
    glMaterialfv( GL_FRONT, GL_AMBIENT_AND_DIFFUSE, sphereDiff);
	glMaterialfv( GL_FRONT, GL_SPECULAR, sphereSpec);
	glMaterialf(  GL_FRONT, GL_SHININESS, 27.);
	glutSolidSphere( 2.5, 500, 500);
    glEndList();

    DivisionsIndex = 5;
    Board = GenerateBoard( Divisions[ DivisionsIndex]);
    
    glEnable( GL_TEXTURE_2D);
    Textures.push_back( LoadTexture( ""));
    Textures.push_back( LoadTexture( "mandrill.bmp"));
    Textures.push_back( LoadTexture( "lena512color.bmp"));
    Textures.push_back( LoadTexture( "stars.bmp"));
    Textures.push_back( LoadTexture( "HDGV.bmp"));
	NoTextureToUse = 0;
    glDisable( GL_TEXTURE_2D);
}


float ObjectManager::DiffuseTextureOn[]  = { 1., 1., 1., 1.},
      ObjectManager::DiffuseTextureOff[] = { 0., 1., 1., 1.};
int   ObjectManager::Divisions[]         = { 1, 5, 10, 50, 100, 250, 500, 1000};


ObjectManager::~ObjectManager()
{
    for ( vector<unsigned int>::const_iterator it=Textures.begin(); it!=Textures.end(); it++)
        if ( glIsTexture( *it))
            glDeleteTextures( 1, &(*it));
}


void ObjectManager::Draw() const
{
    if ( DrawSphere && Sphere != -1 )
        glCallList( Sphere);

    // The board display list never changes, but the texture it uses may change
    if ( ShowingTexture() )
    {
        glMaterialfv( GL_FRONT, GL_AMBIENT_AND_DIFFUSE, DiffuseTextureOn);
        glEnable( GL_TEXTURE_2D);
        glBindTexture( GL_TEXTURE_2D, Textures[ NoTextureToUse]);
        glTexEnvf( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    }
    else
        glMaterialfv( GL_FRONT, GL_AMBIENT_AND_DIFFUSE, DiffuseTextureOff);
    glCallList( Board);
    if ( glIsTexture( Textures[ NoTextureToUse]) == GL_TRUE )
        glDisable( GL_TEXTURE_2D);
}


const bool ObjectManager::ShowSphere()
{
    DrawSphere = !DrawSphere;
    glutPostRedisplay();
    return DrawSphere;
}


const bool ObjectManager::SelectTexture(const int textureNo)
{
    if ( textureNo >= 0 && textureNo < (int) Textures.size() && textureNo != NoTextureToUse )
    {
        NoTextureToUse = textureNo;
        glutPostRedisplay();
        return true;
    }
    return false;
}


void ObjectManager::ChangeNoOfDivisions( const bool increase)
{
    int no = sizeof( Divisions)/sizeof(int);
    int NewDivisionsIndex = min( no-1, max( 0, DivisionsIndex + (increase ? 1: -1)));
    if ( NewDivisionsIndex != DivisionsIndex )
    {
        DivisionsIndex = NewDivisionsIndex;
        glDeleteLists( Board, 1);
        Board = GenerateBoard( Divisions[ DivisionsIndex]);
        glutPostRedisplay();
    }
}


const bool ObjectManager::ShowingTexture() const
{
    // GL_TRUE is a GLboolean that is not the same as bool
    return ( glIsTexture( Textures[ NoTextureToUse]) == GL_TRUE );
}


const unsigned int ObjectManager::LoadTexture( const string & textureFilename)
{
    unsigned int textureID;
    glGenTextures( 1, &textureID);
	BMPloader * bitmap = new BMPloader( textureFilename);
	if ( bitmap->Isvalid() )
	{
		glPixelStoref( GL_UNPACK_ALIGNMENT, 1);
		glEnable( GL_TEXTURE_2D);
        glBindTexture( GL_TEXTURE_2D, textureID);
		glTexImage2D( GL_TEXTURE_2D, 0, GL_RGB, bitmap->Width(), bitmap->Heigth(), 0, GL_RGB,
					 GL_UNSIGNED_BYTE, bitmap->Image());
		glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
		glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
		glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
		glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	}
	delete bitmap;
	return textureID;
}


const unsigned int ObjectManager::GenerateBoard( const int NoDivisions)
{
    unsigned int board = glGenLists( 1);
    float materialDiffuse[] = {.54, .89, .63, 1.};
    glNewList( board, GL_COMPILE);
    glMaterialfv( GL_FRONT, GL_AMBIENT_AND_DIFFUSE, materialDiffuse);
    glPushMatrix(),
    glTranslatef( 0., 0., -2.5);
    glNormal3f( 0., 0., 1.);
    float x1, x2, y1, y2, s1, s2, t1, t2;
    y2 = -5;   t2 = 0;
    for ( int j=0; j<NoDivisions; j++) {
        y1 = y2;   t1 = t2;
        y2 = 10. * (j+1) / NoDivisions - 10. * .5;
        t2 = 1. * (j+1) / NoDivisions;
        x2 = -5.;   s2 = 0.;
        for ( int i=0; i<NoDivisions; i++) {
            x1 = x2;   s1 = s2;
            x2 = 10. * (i+1) / NoDivisions - 10 * .5;
            s2 = 1. * (i+1) / NoDivisions;
            glBegin( GL_TRIANGLE_STRIP);
            glTexCoord2f( s1, t1);
            glVertex3f(   x1, y1, 0.);
            glTexCoord2f( s2, t1);
            glVertex3f(   x2, y1, 0.);
            glTexCoord2f( s1, t2);
            glVertex3f(   x1, y2, 0.);
            glTexCoord2f( s2, t2);
            glVertex3f(   x2, y2, 0.);
            glEnd();
        }
    }
    glPopMatrix();
    glEndList();
    
    return board;
}


// Camera --------------------------------------------------------------------------


Camera::Camera()
{
    SetProjection();
	glLoadIdentity();
	gluLookAt( .5, -10., 10., .5, 0., -2.5, 0., 0., 1.);
}


void Camera::SetProjection()
{
    // Called whenever the window is resized
    glMatrixMode( GL_PROJECTION);
    glLoadIdentity();
	gluPerspective( 45., ((float) glutGet( GLUT_WINDOW_WIDTH)) / glutGet( GLUT_WINDOW_HEIGHT),
                   .001, 1000.);
    glMatrixMode( GL_MODELVIEW);
}


// LightSource -------------------------------------------------------------------


LightSource::LightSource()
{
    LightAngle  = 90;
    SpotlightOn = false;
    AmbientOn   = false;
    
    float _difspec[] = {  1.,  1.,  1.,  1.};
    
    glLightfv( GL_LIGHT0, GL_DIFFUSE,  _difspec);
    glLightfv( GL_LIGHT0, GL_SPECULAR, _difspec);
    glLightf(  GL_LIGHT0, GL_SPOT_CUTOFF, 35.);
    glLightf(  GL_LIGHT0, GL_SPOT_EXPONENT, 1.);
    
    glLightfv( GL_LIGHT1, GL_DIFFUSE,  _difspec);
    glLightfv( GL_LIGHT1, GL_SPECULAR, _difspec);
    
    ToggleLightSource();
    ToggleAmbientLight();
}


int   LightSource::AngleIncrement =   5,
      LightSource::MaxAngle       = 180,
      LightSource::MinAngle       =   0;
float LightSource::LightPosition[]  = { 0., -4., 0., 0.},
      LightSource::LightDirection[] = { 0., 1., 0., 1.},
      LightSource::AmbientModel[]   = { .3, .3, .3, 1.},
      LightSource::NoAmbientModel[] = { 0., 0., 0., 1.};


void LightSource::Locate() const
{
	glPushMatrix();
	glRotatef( 35., 0., 0., 1.);
	glRotatef( -LightAngle, 1., 0., 0.);
	if ( SpotlightOn )
	{
		glLightfv( GL_LIGHT0, GL_POSITION,       LightPosition);
		glLightfv( GL_LIGHT0, GL_SPOT_DIRECTION, LightDirection);
	}
	else
		glLightfv( GL_LIGHT1, GL_POSITION, LightPosition);
	glPopMatrix();
}


const bool LightSource::ToggleLightSource()
{
    SpotlightOn = !SpotlightOn;
    if ( SpotlightOn )
    {
        glEnable( GL_LIGHT0);
        glDisable( GL_LIGHT1);
    }
    else
    {
        glEnable( GL_LIGHT1);
        glDisable( GL_LIGHT0);
    }
    LightPosition[3] = SpotlightOn ? 1. : 0.;
    glutPostRedisplay();
    return SpotlightOn;
}


const bool LightSource::SetIncidenceAngle( const bool increase)
{
    int angle = (increase ? AngleIncrement : -AngleIncrement) + LightAngle;
    angle = min( MaxAngle, max( angle, MinAngle));
    if ( angle != LightAngle )
    {
        LightAngle = angle;
        glutPostRedisplay();
        return true;
    }
    return false;
}


const bool LightSource::ToggleAmbientLight()
{
    AmbientOn = !AmbientOn;
    glLightModelfv( GL_LIGHT_MODEL_AMBIENT, AmbientOn ? AmbientModel : NoAmbientModel);
    glutPostRedisplay();
    return AmbientOn;
}


// OpenGLApplication --------------------------------------------------------------------------


OpenGLApplication::OpenGLApplication()
{
    HUDon        = true;
    FullScreenOn = false;
}


OpenGLApplication OpenGLApplication::MyApplication;

string OpenGLApplication::label1 = "Arrow keys-move light source;   -/+ increase/decrease no tiles;",
       OpenGLApplication::label2 = "p-parallel/spot;   a-ambient on/off;   s-show/hide sphere",
       OpenGLApplication::label3 = "1-5: select texture; Esc to exit";


OpenGLApplication::~OpenGLApplication()
{
    if ( MyLightSource != NULL )
        delete MyLightSource;
    if ( MyCamera != NULL )
        delete MyCamera;
    if ( MyObjects != NULL )
        delete MyObjects;
}


void OpenGLApplication::Initialize(int argc, char **argv)
{
    glutInit( &argc, argv);
	glutInitDisplayMode( GLUT_SINGLE | GLUT_RGB | GLUT_DEPTH);
	glutInitWindowSize( 640,480);
    glutCreateWindow( "Parallel/Spot Light, Tiles and Textures");
    
    glClearColor( 0., 0., 0., 0.);
    MyLightSource = new LightSource();
    MyObjects     = new ObjectManager();
    MyCamera      = new Camera();
    
    glutReshapeFunc( reshape);
    glutDisplayFunc( display);
    glutKeyboardFunc( keyboard);
    glutSpecialFunc( special);
    
	glEnable( GL_DEPTH_TEST);
	glEnable( GL_LIGHTING);
    glEnable( GL_NORMALIZE);
    
    glutMainLoop();
}


void OpenGLApplication::MyDisplay()
{
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    
    MyLightSource->Locate();
    MyObjects->Draw();
    
    if ( HUDon )
        HUD();
    
    glFlush();
}


void OpenGLApplication::MyReshape( const int w, const int h)
{
    glViewport( 0, 0, w, h);
    MyCamera->SetProjection();
}


void OpenGLApplication::MyKeyboard( const unsigned char key, const int x, const int y)
{
    switch ( tolower( key) ) {
        case 27:
            exit( 0);
            break;
        case 'p':
            MyLightSource->ToggleLightSource();
            break;
        case 'a':
            MyLightSource->ToggleAmbientLight();
            break;
        case 's':
            MyObjects->ShowSphere();
            break;
        case '1':
        case '2':
        case '3':
        case '4':
        case '5':
        case '6':
        case '7':
        case '8':
        case '9':
            MyObjects->SelectTexture( key - '1');
            break;
        case '+':
        case '-':
            MyObjects->ChangeNoOfDivisions( key == '+');
            break;
        case 'f':
            FullScreenOn = !FullScreenOn;
            if ( FullScreenOn )
                glutFullScreen();
            else
            {
                glutReshapeWindow( glutGet( GLUT_INIT_WINDOW_WIDTH), glutGet( GLUT_INIT_WINDOW_HEIGHT));
                glutPositionWindow( glutGet( GLUT_INIT_WINDOW_X), glutGet( GLUT_INIT_WINDOW_Y));
            }
            break;
        case 'h':
            HUDon = !HUDon;
            glutPostRedisplay();
            break;
        default:
            break;
    }
}


void OpenGLApplication::MySpecialFunction( const int key, const int x, const int y)
{
    switch ( key ) {
		case GLUT_KEY_RIGHT:
		case GLUT_KEY_DOWN:
			MyLightSource->SetIncidenceAngle( false);
			break;
		case GLUT_KEY_LEFT:
		case GLUT_KEY_UP:
			MyLightSource->SetIncidenceAngle( true);
			break;
        default:
            break;
    }
}


void OpenGLApplication::HUD()
{
    // Switch to raster mode (saving active projection)
    glDisable( GL_LIGHTING);
    glMatrixMode( GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();
    gluOrtho2D( 0, glutGet( GLUT_WINDOW_WIDTH), 0, glutGet( GLUT_WINDOW_HEIGHT));
    glMatrixMode( GL_MODELVIEW);
    glPushMatrix();
    glLoadIdentity();
    
    glColor3f( 0., 1., 0.);
    DisplayString( 5, 30, label1);
    DisplayString( 5, 18, label2);
    DisplayString( 5,  5, label3);
    
	ostringstream status;
	status << ( MyLightSource->IsSpotOn() ? "Spot" : "Parallel") << " light  -  Ambient "
           << ( MyLightSource->IsAmbientOn() ? "on" : "off") << "  -  texture "
           << (MyObjects->ShowingTexture() ? "on" : "off");
	DisplayString( 5, glutGet( GLUT_WINDOW_HEIGHT) - 18, status.str(), GLUT_BITMAP_HELVETICA_18);
	status.str( "");
	status << "Light source at " << abs(MyLightSource->LightSourceAngle()) << " deg to plane";
	DisplayString( 5, glutGet( GLUT_WINDOW_HEIGHT)-36, status.str(), GLUT_BITMAP_HELVETICA_18);
	status.str( "");
	status << "Tiling: " << MyObjects->NoOfDivisions() << " x " << MyObjects->NoOfDivisions();
	DisplayString( 5, glutGet( GLUT_WINDOW_HEIGHT)-54, status.str(), GLUT_BITMAP_HELVETICA_18);
    
    // Switch back to normal mode (restoring projection)
    glMatrixMode( GL_PROJECTION);
    glPopMatrix();
    glMatrixMode( GL_MODELVIEW);
    glPopMatrix();
    glEnable( GL_LIGHTING);
}


void OpenGLApplication::DisplayString( const int x, const int y, const string &label, const void * font)
{
    void * fontToUse = (font != NULL) ? (void*) font : GLUT_BITMAP_HELVETICA_12;
    glRasterPos2i( x, y);
    for ( int i=0; i<(int)label.size(); i++)
        glutBitmapCharacter( fontToUse, label[i]);
}




// Main - to start it all ------------------------------------------------------------


int main( int argc, char ** argv)
{
    OpenGLApplication::MyApplication.Initialize( argc, argv);
    return 0;
}